from os import path
from pytest import fixture
from yaml import dump


@fixture(scope='function')
def cfg_path(postgres_dsn, session_manager, temporary_directory, redis_url):
    cfg = {
        'applications': [
            {
                'path': '/foo/*',
                'application': 'onegov.core.Framework',
                'namespace': 'foo',
                'configuration': {
                    'dsn': postgres_dsn,
                    'redis_url': redis_url
                }
            }
        ]
    }

    session_manager.ensure_schema_exists('foo-bar')

    cfg_path = path.join(temporary_directory, 'onegov.yml')
    with open(cfg_path, 'w') as f:
        f.write(dump(cfg))

    return cfg_path



import pytest
import transaction

from datetime import datetime
from datetime import timedelta
from onegov.event import Event
from onegov.event import Occurrence
from onegov.gis import Coordinates
from sedate import replace_timezone


def tzdatetime(year, month, day, hour, minute, timezone):
    return replace_timezone(datetime(year, month, day, hour, minute), timezone)


def test_transitions():

    event = Event(state='initiated')
    assert event.state == 'initiated'

    with pytest.raises(AssertionError):
        event.publish()
    with pytest.raises(AssertionError):
        event.withdraw()
    event.submit()
    assert event.state == 'submitted'

    with pytest.raises(AssertionError):
        event.submit()
    with pytest.raises(AssertionError):
        event.withdraw()
    event.publish()
    assert event.state == 'published'

    with pytest.raises(AssertionError):
        event.submit()
    with pytest.raises(AssertionError):
        event.publish()
    event.withdraw()
    assert event.state == 'withdrawn'

    with pytest.raises(AssertionError):
        event.submit()
    with pytest.raises(AssertionError):
        event.withdraw()
    event.publish()
    assert event.state == 'published'


def test_create_event(session):
    timezone = 'Europe/Zurich'
    start = tzdatetime(2008, 2, 7, 10, 15, timezone)
    end = tzdatetime(2008, 2, 7, 16, 00, timezone)
    title = 'Squirrel Park Visit'
    description = '<em>Furry</em> things will happen!'
    location = 'Squirrel Park'
    tags = ['fun', 'animals', 'furry']

    event = Event(state='initiated')
    event.timezone = timezone
    event.start = start
    event.end = end
    event.title = title
    event.content = {'description': description}
    event.location = location
    event.tags = tags
    event.meta = {'submitter': 'fat.pauly@squirrelpark.org'}
    event.name = 'event'
    session.add(event)

    event.submit()
    event.publish()

    transaction.commit()

    event = session.query(Event).one()
    assert event.timezone == timezone
    assert event.start == start
    assert event.localized_start == start
    assert str(event.start.tzinfo) == 'UTC'
    assert str(event.localized_start.tzinfo) == timezone
    assert event.end == end
    assert event.localized_end == end
    assert str(event.end.tzinfo) == 'UTC'
    assert str(event.localized_end.tzinfo) == timezone
    assert event.title == title
    assert event.location == location
    assert sorted(event.tags) == sorted(tags)
    assert event.description == description
    assert event.content == {'description': description}
    assert event.meta == {'submitter': 'fat.pauly@squirrelpark.org'}
    assert event.source is None
    assert event.name == 'event'

    occurrence = session.query(Occurrence).one()
    assert occurrence.timezone == timezone
    assert occurrence.start == start
    assert occurrence.localized_start == start
    assert str(occurrence.start.tzinfo) == 'UTC'
    assert str(occurrence.localized_start.tzinfo) == timezone
    assert occurrence.end == end
    assert occurrence.localized_end == end
    assert str(occurrence.end.tzinfo) == 'UTC'
    assert str(occurrence.localized_end.tzinfo) == timezone
    assert occurrence.title == title
    assert occurrence.location == location
    assert sorted(occurrence.tags) == sorted(tags)
    assert occurrence.event.description == description
    assert occurrence.name == 'event-2008-02-07'

    assert [o.id for o in event.occurrences] == [occurrence.id]
    assert occurrence.event.id == event.id


def test_icalendar_recurrence():
    event = Event(state='initiated')
    assert event.icalendar_recurrence == None

    event.recurrence = (
        'RRULE:FREQ=WEEKLY;'
        'UNTIL=20150616T220000Z;'
        'BYDAY=MO,WE,FR'
    )
    assert event.icalendar_recurrence['FREQ'] == ['WEEKLY']
    assert event.icalendar_recurrence['UNTIL'][0].year == 2015
    assert event.icalendar_recurrence['UNTIL'][0].month == 6
    assert event.icalendar_recurrence['UNTIL'][0].day == 16
    assert event.icalendar_recurrence['BYDAY'] == ['MO', 'WE', 'FR']


def test_occurrence_dates(session):
    year = datetime.today().year

    event = Event(state='initiated')
    event.timezone = 'Europe/Zurich'
    event.start = tzdatetime(year, 2, 7, 10, 15, 'Europe/Zurich')
    event.end = tzdatetime(year, 2, 7, 16, 00, 'Europe/Zurich')
    event.recurrence = (
        f'RRULE:FREQ=WEEKLY;'
        f'UNTIL={year+2}0211T100000Z;'
        f'BYDAY=MO,TU,WE,TH,FR,SA,SU'
    )

    assert len(event.occurrence_dates(limit=False)) > 700

    dates = event.occurrence_dates()
    assert len(dates) < 700
    assert dates[0] == tzdatetime(year, 2, 7, 10, 15, 'Europe/Zurich')
    assert dates[-1] == tzdatetime(year + 1, 12, 31, 10, 15, 'Europe/Zurich')
    assert str(dates[0].tzinfo) == 'UTC'
    assert str(dates[-1].tzinfo) == 'UTC'

    dates = event.occurrence_dates(localize=True)
    assert len(dates) < 700
    assert dates[0] == tzdatetime(year, 2, 7, 10, 15, 'Europe/Zurich')
    assert dates[-1] == tzdatetime(year + 1, 12, 31, 10, 15, 'Europe/Zurich')
    assert str(dates[0].tzinfo) == 'Europe/Zurich'
    assert str(dates[-1].tzinfo) == 'Europe/Zurich'

    event.title = 'Event'
    session.add(event)
    transaction.commit()

    event = session.query(Event).one()
    assert len(event.occurrence_dates(limit=False)) > 700

    dates = event.occurrence_dates()
    assert len(dates) < 700
    assert dates[0] == tzdatetime(year, 2, 7, 10, 15, 'Europe/Zurich')
    assert dates[-1] == tzdatetime(year + 1, 12, 31, 10, 15, 'Europe/Zurich')
    assert str(dates[0].tzinfo) == 'UTC'
    assert str(dates[-1].tzinfo) == 'UTC'

    dates = event.occurrence_dates(localize=True)
    assert len(dates) < 700
    assert dates[0] == tzdatetime(year, 2, 7, 10, 15, 'Europe/Zurich')
    assert dates[-1] == tzdatetime(year + 1, 12, 31, 10, 15, 'Europe/Zurich')
    assert str(dates[0].tzinfo) == 'Europe/Zurich'
    assert str(dates[-1].tzinfo) == 'Europe/Zurich'


def test_latest_occurrence(session):

    def create_event(delta):
        start = datetime.now() + delta
        end = start + timedelta(hours=6)
        session.query(Occurrence).delete()
        session.query(Event).delete()
        session.add(
            Event(
                state='published',
                title='Event',
                timezone='Europe/Zurich',
                start=replace_timezone(start, 'Europe/Zurich'),
                end=replace_timezone(end, 'Europe/Zurich'),
                recurrence=(
                    f'RRULE:FREQ=WEEKLY;'
                    f'UNTIL={end.year}{end.month:02}{end.day:02}T220000Z;'
                    f'BYDAY=MO,TU,WE,TH,FR,SA,SU'
                ),
            )
        )
        transaction.commit()
        return session.query(Event).one()

    # current
    event = create_event(timedelta(hours=-3))
    assert event.latest_occurrence.start == event.occurrence_dates()[0]

    event = create_event(timedelta(days=-1, hours=-3))
    assert event.latest_occurrence.start == event.occurrence_dates()[1]

    # past
    event = create_event(timedelta(days=-40))
    assert event.latest_occurrence.start == event.occurrence_dates()[-1]

    # future
    event = create_event(timedelta(days=40))
    assert event.latest_occurrence.start == event.occurrence_dates()[0]


def test_occurrence_dates_dst(session):
    year = datetime.today().year

    event = Event(state='initiated')
    event.timezone = 'Europe/Zurich'
    event.start = tzdatetime(year, 1, 1, 10, 15, 'Europe/Zurich')
    event.end = tzdatetime(year, 1, 11, 16, 00, 'Europe/Zurich')

    with pytest.raises(RuntimeError) as e:
        event.recurrence = (
            f'RRULE:FREQ=WEEKLY;'
            f'UNTIL={year}12310000;'
            f'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        )

    assert 'UNTIL is not timezone-aware' in str(e)

    event.recurrence = (
        f'RRULE:FREQ=WEEKLY;'
        f'UNTIL={year}12310000Z;'
        f'BYDAY=MO,TU,WE,TH,FR,SA,SU'
    )
    occurrences = event.occurrence_dates(localize=True)
    assert all((o.hour == 10 for o in occurrences))


def test_create_event_recurring(session):
    timezone = 'Europe/Zurich'
    start = tzdatetime(2008, 2, 7, 10, 15, timezone)
    end = tzdatetime(2008, 2, 7, 16, 00, timezone)
    title = 'Squirrel Park Visit'
    description = '<em>Furry</em> things will happen!'
    location = 'Squirrel Park'
    tags = ['fun', 'animals', 'furry']

    event = Event(state='initiated')
    event.timezone = timezone
    event.start = start
    event.end = end
    event.recurrence = (
        'RRULE:FREQ=WEEKLY;'
        'UNTIL=200802111500Z;'
        'BYDAY=MO,TU,WE,TH,FR,SA,SU'
    )
    event.title = title
    event.content = {'description': description}
    event.location = location
    event.tags = tags
    event.meta = {'submitter': 'fat.pauly@squirrelpark.org'}
    event.name = 'event'
    session.add(event)

    event.submit()
    event.publish()

    transaction.commit()

    event = session.query(Event).one()
    assert event.timezone == timezone
    assert event.start == start
    assert event.localized_start == start
    assert str(event.start.tzinfo) == 'UTC'
    assert str(event.localized_start.tzinfo) == timezone
    assert event.end == end
    assert event.localized_end == end
    assert str(event.end.tzinfo) == 'UTC'
    assert str(event.localized_end.tzinfo) == timezone
    assert event.recurrence == (
        'RRULE:FREQ=WEEKLY;'
        'UNTIL=200802111500Z;'
        'BYDAY=MO,TU,WE,TH,FR,SA,SU'
    )
    assert event.title == title
    assert event.location == location
    assert sorted(event.tags) == sorted(tags)
    assert event.description == description
    assert event.content == {'description': description}
    assert event.meta == {'submitter': 'fat.pauly@squirrelpark.org'}
    assert event.name == 'event'

    occurrences = session.query(Occurrence).all()
    assert len(occurrences) == 5
    for occurrence in occurrences:
        assert occurrence.timezone == timezone
        assert str(occurrence.start.tzinfo) == 'UTC'
        assert str(occurrence.localized_start.tzinfo) == timezone
        assert str(occurrence.end.tzinfo) == 'UTC'
        assert str(occurrence.localized_end.tzinfo) == timezone
        assert occurrence.title == title
        assert occurrence.location == location
        assert sorted(occurrence.tags) == sorted(tags)
        assert occurrence.event.description == description

    assert occurrences[0].start == start
    assert occurrences[0].localized_start == start
    assert occurrences[0].end == end
    assert occurrences[0].localized_end == end
    assert occurrences[0].name == 'event-2008-02-07'

    start = tzdatetime(2008, 2, 8, 10, 15, timezone)
    end = tzdatetime(2008, 2, 8, 16, 00, timezone)
    assert occurrences[1].start == start
    assert occurrences[1].localized_start == start
    assert occurrences[1].end == end
    assert occurrences[1].localized_end == end
    assert occurrences[1].name == 'event-2008-02-08'

    start = tzdatetime(2008, 2, 9, 10, 15, timezone)
    end = tzdatetime(2008, 2, 9, 16, 00, timezone)
    assert occurrences[2].start == start
    assert occurrences[2].localized_start == start
    assert occurrences[2].end == end
    assert occurrences[2].localized_end == end
    assert occurrences[2].name == 'event-2008-02-09'

    start = tzdatetime(2008, 2, 10, 10, 15, timezone)
    end = tzdatetime(2008, 2, 10, 16, 00, timezone)
    assert occurrences[3].start == start
    assert occurrences[3].localized_start == start
    assert occurrences[3].end == end
    assert occurrences[3].localized_end == end
    assert occurrences[3].name == 'event-2008-02-10'

    start = tzdatetime(2008, 2, 11, 10, 15, timezone)
    end = tzdatetime(2008, 2, 11, 16, 00, timezone)
    assert occurrences[4].start == start
    assert occurrences[4].localized_start == start
    assert occurrences[4].end == end
    assert occurrences[4].localized_end == end
    assert occurrences[4].name == 'event-2008-02-11'

    assert (sorted([o.id for o in event.occurrences]) ==
            sorted([o.id for o in occurrences]))
    assert occurrences[0].event.id == event.id
    assert occurrences[1].event.id == event.id
    assert occurrences[2].event.id == event.id
    assert occurrences[3].event.id == event.id
    assert occurrences[4].event.id == event.id


def test_update_event(session):
    timezone = 'Europe/Zurich'

    event = Event(state='initiated')
    event.timezone = timezone
    event.start = tzdatetime(2008, 2, 7, 10, 15, timezone)
    event.end = tzdatetime(2008, 2, 7, 16, 00, timezone)
    event.title = 'Squirrel Park Visit'
    event.content = {'description': '<em>Furry</em> things will happen!'}
    event.location = 'Squirrel Park'
    event.tags = ['fun', 'animals', 'furry']
    event.name = 'event'
    session.add(event)

    event.submit()
    event.publish()

    occurence = session.query(Event).one().occurrences[0]
    assert occurence.timezone == event.timezone
    assert occurence.start == event.start
    assert occurence.end == event.end
    assert occurence.event.title == event.title
    assert occurence.title == event.title
    assert occurence.event.content == event.content
    assert occurence.event.description == event.content['description']
    assert occurence.event.location == event.location
    assert occurence.location == event.location
    assert sorted(occurence.event.tags) == sorted(event.tags)
    assert sorted(occurence.tags) == sorted(event.tags)
    assert occurence.name == 'event-2008-02-07'

    event.start = tzdatetime(2009, 2, 7, 10, 15, timezone)
    event.end = tzdatetime(2009, 2, 7, 10, 15, timezone)
    event.title = 'Squirrel Park Visit - Cancelled'
    event.content = {'description': 'No <em>Furry</em> things will happen! :('}
    event.location = '-'
    event.tags = ['no fun', 'no animals']

    assert len(session.query(Event).one().occurrences) == 1
    occurence = session.query(Event).one().occurrences[0]
    assert occurence.timezone == event.timezone
    assert occurence.start == event.start
    assert occurence.end == event.end
    assert occurence.event.title == event.title
    assert occurence.title == event.title
    assert occurence.event.content == event.content
    assert occurence.event.description == event.content['description']
    assert occurence.event.location == event.location
    assert occurence.location == event.location
    assert sorted(occurence.event.tags) == sorted(event.tags)
    assert sorted(occurence.tags) == sorted(event.tags)

    event.recurrence = (
        'RRULE:FREQ=WEEKLY;'
        'UNTIL=20090209T090000Z;'
        'BYDAY=MO,TU,WE,TH,FR,SA,SU'
    )
    occurrences = session.query(Event).one().occurrences
    assert len(occurrences) == 2
    assert occurrences[0].start == tzdatetime(2009, 2, 7, 10, 15, timezone)
    assert occurrences[0].end == tzdatetime(2009, 2, 7, 10, 15, timezone)
    assert occurrences[1].start == tzdatetime(2009, 2, 8, 10, 15, timezone)
    assert occurrences[1].end == tzdatetime(2009, 2, 8, 10, 15, timezone)

    event.timezone = 'US/Eastern'
    assert session.query(Event).one().timezone == 'US/Eastern'

    occurrences = session.query(Event).one().occurrences
    assert len(occurrences) == 2
    assert occurrences[0].start == tzdatetime(2009, 2, 7, 10, 15, timezone)
    assert occurrences[0].end == tzdatetime(2009, 2, 7, 10, 15, timezone)
    assert occurrences[0].timezone == 'US/Eastern'
    assert occurrences[1].start == tzdatetime(2009, 2, 8, 10, 15, timezone)
    assert occurrences[1].end == tzdatetime(2009, 2, 8, 10, 15, timezone)
    assert occurrences[1].timezone == 'US/Eastern'

    event.name = 'new-event'
    assert session.query(Event).one().name == 'new-event'

    occurrences = session.query(Event).one().occurrences
    assert len(occurrences) == 2
    assert occurrences[0].name == 'new-event-2009-02-07'
    assert occurrences[1].name == 'new-event-2009-02-08'

    event.recurrence = ''
    assert session.query(Occurrence).count() == 1
    assert len(session.query(Event).one().occurrences) == 1


def test_delete_event(session):
    timezone = 'Europe/Zurich'
    event = Event(
        state='initiated',
        title='event',
        start=tzdatetime(2008, 2, 7, 10, 15, timezone),
        end=tzdatetime(2008, 2, 7, 16, 00, timezone),
        timezone=timezone
    )
    session.add(event)
    session.delete(session.query(Event).one())
    transaction.commit()
    session.expire_all()
    assert session.query(Event).count() == 0
    assert session.query(Occurrence).count() == 0

    event = Event(
        state='initiated',
        title='event',
        start=tzdatetime(2008, 2, 7, 10, 15, timezone),
        end=tzdatetime(2008, 2, 7, 16, 00, timezone),
        timezone=timezone
    )
    session.add(event)
    event.submit()
    session.delete(session.query(Event).one())
    transaction.commit()
    session.expire_all()
    assert session.query(Event).count() == 0
    assert session.query(Occurrence).count() == 0

    event = Event(
        state='initiated',
        title='event',
        start=tzdatetime(2008, 2, 7, 10, 15, timezone),
        end=tzdatetime(2008, 2, 7, 16, 00, timezone),
        timezone=timezone
    )
    session.add(event)
    event.submit()
    event.publish()
    session.delete(session.query(Event).one())
    transaction.commit()
    session.expire_all()
    assert session.query(Event).count() == 0
    assert session.query(Occurrence).count() == 0

    event = Event(
        state='initiated',
        title='event',
        start=tzdatetime(2008, 2, 7, 10, 15, timezone),
        end=tzdatetime(2008, 2, 7, 16, 00, timezone),
        timezone=timezone
    )
    session.add(event)
    event.submit()
    event.publish()
    event.withdraw()
    session.delete(session.query(Event).one())
    transaction.commit()
    session.expire_all()
    assert session.query(Event).count() == 0
    assert session.query(Occurrence).count() == 0


def test_as_ical():
    url = 'https://example.org/my-event'
    event = Event(
        state='initiated',
        timezone='Europe/Zurich',
        start=tzdatetime(2008, 2, 7, 10, 15, 'Europe/Zurich'),
        end=tzdatetime(2008, 2, 7, 16, 00, 'Europe/Zurich'),
        recurrence=(
            'RRULE:FREQ=WEEKLY;'
            'UNTIL=20080207T150000Z;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        ),
        title='Squirrel Park Visit',
        content={'description': '<em>Furry</em> things will happen!'},
        location='Squirrel Park',
        tags=['fun', 'animals', 'furry'],
        meta={'submitter': 'fat.pauly@squirrelpark.org'},
        name='event',
        modified=tzdatetime(2008, 2, 7, 10, 15, 'Europe/Zurich'),
        coordinates=Coordinates(47.051752750515746, 8.305739625357093)
    )
    ical = event.as_ical(url=url).decode().strip().split('\r\n')
    assert sorted(ical) == sorted([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'UID:event@onegov.event',
        'SUMMARY:Squirrel Park Visit',
        'DTSTART;VALUE=DATE-TIME:20080207T091500Z',
        'DTEND;VALUE=DATE-TIME:20080207T150000Z',
        'DTSTAMP;VALUE=DATE-TIME:20080207T091500Z',
        'DESCRIPTION:<em>Furry</em> things will happen!',
        'CATEGORIES:fun,animals,furry',
        'LAST-MODIFIED;VALUE=DATE-TIME:20080207T091500Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        (
            'RRULE:FREQ=WEEKLY;'
            'UNTIL=20080207T150000Z;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        ),
        'URL:https://example.org/my-event',
        'END:VEVENT',
        'END:VCALENDAR',
    ])

    event.submit()
    event.publish()
    occurrence = event.occurrences[0]
    occurrence.modified = tzdatetime(2008, 2, 7, 10, 15, 'Europe/Zurich')

    ical = event.occurrences[0].as_ical(url=url).decode().strip().split('\r\n')
    assert sorted(ical) == sorted([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'UID:event-2008-02-07@onegov.event',
        'SUMMARY:Squirrel Park Visit',
        'DTSTART;VALUE=DATE-TIME:20080207T091500Z',
        'DTEND;VALUE=DATE-TIME:20080207T150000Z',
        'DTSTAMP;VALUE=DATE-TIME:20080207T091500Z',
        'DESCRIPTION:<em>Furry</em> things will happen!',
        'CATEGORIES:fun,animals,furry',
        'LAST-MODIFIED;VALUE=DATE-TIME:20080207T091500Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/my-event',
        'END:VEVENT',
        'END:VCALENDAR',
    ])



from datetime import datetime
from dateutil import rrule
from dateutil.rrule import rrulestr
from icalendar import Calendar as vCalendar
from icalendar import Event as vEvent
from icalendar import vRecur
from onegov.core.orm import Base
from onegov.core.orm.abstract import associated
from onegov.core.orm.mixins import content_property
from onegov.core.orm.mixins import ContentMixin
from onegov.core.orm.mixins import meta_property
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.event.models.mixins import OccurrenceMixin
from onegov.event.models.occurrence import Occurrence
from onegov.file import File
from onegov.gis import CoordinatesMixin
from onegov.search import ORMSearchable
from pytz import UTC
from sedate import standardize_date
from sedate import to_timezone
from sqlalchemy import and_
from sqlalchemy import Column
from sqlalchemy import desc
from sqlalchemy import Enum
from sqlalchemy import func
from sqlalchemy import Text
from sqlalchemy.orm import backref
from sqlalchemy.orm import object_session
from sqlalchemy.orm import relationship
from sqlalchemy.orm import validates
from uuid import uuid4


class EventFile(File):
    __mapper_args__ = {'polymorphic_identity': 'eventfile'}


class Event(Base, OccurrenceMixin, ContentMixin, TimestampMixin,
            ORMSearchable, CoordinatesMixin):
    """ Defines an event.

    Occurrences are stored in a seperate table containing only a minimal set
    of attributes from the event. This could also be archieved using postgres
    directly with dateutil/plpythonu/pg_rrule and materialized views.

    Occurrences are only created/updated, if the event is published.
    Occurrences are created only for this and the next year.
    """

    __tablename__ = 'events'

    #: Internal number of the event
    id = Column(UUID, primary_key=True, default=uuid4)

    #: State of the event
    state = Column(
        Enum('initiated', 'submitted', 'published', 'withdrawn',
             name='event_state'),
        nullable=False,
        default='initiated'
    )

    #: description of the event
    description = content_property()

    #: the event organizer
    organizer = content_property()

    #: the source of the event, if imported
    source = meta_property()

    #: Recurrence of the event (RRULE, see RFC2445)
    recurrence = Column(Text, nullable=True)

    #: The associated image
    image = associated(EventFile, 'image', 'one-to-one', uselist=False)

    #: Recurrence of the event as icalendar object
    @property
    def icalendar_recurrence(self):
        if not self.recurrence:
            return None
        return vRecur.from_ical(self.recurrence.replace('RRULE:', ''))

    #: Occurences of the event
    occurrences = relationship(
        "Occurrence",
        cascade="all, delete-orphan",
        backref=backref("event"),
        lazy='joined',
    )

    es_properties = {
        'title': {'type': 'localized'},
        'description': {'type': 'localized'},
        'location': {'type': 'localized'},
        'organizer': {'type': 'localized'}
    }

    @property
    def es_public(self):
        return self.state == 'published'

    @property
    def es_skip(self):
        return self.state == 'submitted' or self.state == 'withdrawn'

    def source_url(self, request):
        """ Returns an url pointing to the external event if imported. """
        if not self.source:
            return None

        if self.source.startswith('guidle'):
            guidle_id = self.source.split('-')[-1].split('.')[0]

            return f"https://www.guidle.com/angebote/{guidle_id}"

    def __setattr__(self, name, value):
        """ Automatically update the occurrences if shared attributes change.
        """

        super().__setattr__(name, value)
        if name in ('state', 'title', 'name', 'location', 'tags',
                    'start', 'end', 'timezone', 'recurrence'):
            self._update_occurrences()

    @property
    def latest_occurrence(self):
        """ Returns the occurrence which is presently occurring, the next
        one to occur or the last occurrence.

        """

        session = object_session(self)

        base = session.query(Occurrence).filter_by(event_id=self.id)

        current = base.filter(and_(
            Occurrence.start <= func.now(),
            Occurrence.end >= func.now()
        )).order_by(Occurrence.start).limit(1)

        future = base.filter(
            Occurrence.start >= func.now()
        ).order_by(Occurrence.start).limit(1)

        past = base.filter(
            Occurrence.end <= func.now()
        ).order_by(desc(Occurrence.start))

        return current.union_all(future, past).first()

    @validates('recurrence')
    def validate_recurrence(self, key, r):
        """ Our rrules are quite limited in their complexity. This validator
        makes sure that is actually the case.

        This is a somewhat harsh limit, but it mirrors the actual use of
        onegov.event at this point. More complex rrules are not handled by the
        UI, nor is there currently a plan to do so.

        Currently supported are weekly recurrences and lists of rdates.

        The rational is that people commonly add recurring events on a weekly
        basis (which is a lot of work for a whole year). Or on a monthly
        or yearly basis, in which case selection of single dates is
        acceptable, or even preferrable to complex rrules.

        This UI talk doesn't belong into a module of course, but it is again
        a reailty that only a strict subset of rules is handled and so we want
        to catch events which we cannot edit in our UI early if they are
        imported from outside.

        """
        if r:
            rule = rrulestr(r)

            # a rule must either have a frequency or be a list of rdates
            if not hasattr(rule, '_freq'):
                if all((l.startswith('RDATE') for l in r.splitlines())):
                    return r

                raise RuntimeError(f"'{r}' is too complex")

            # we also only do weekly recurrences (they can also be used
            # to do daily recurrences if they are set to include all days)
            if not rule._freq == rrule.WEEKLY:
                raise RuntimeError(f"The frequency of '{r}' is not WEEKLY")

            # we require a definite end
            if not getattr(rule, '_until'):
                raise RuntimeError(f"'{r}' has no UNTIL")

            # we also want the end date to be timezone-aware
            if rule._until.tzinfo is None:
                raise RuntimeError(f"'{r}''s UNTIL is not timezone-aware")

        return r

    def occurrence_dates(self, limit=True, localize=False):
        """ Returns the start dates of all occurrences.

        Returns non-localized dates per default. Limits the occurrences per
        default to this and the next year.
        """

        def to_local(dt, timezone):
            if dt.tzinfo:
                return to_timezone(dt, timezone).replace(tzinfo=None)
            return dt

        dates = [self.start]
        if self.recurrence:
            # Make sure the RRULE uses local dates (or else the DST is wrong)
            start_local = to_local(self.start, self.timezone)
            try:
                rule = rrulestr(self.recurrence, dtstart=self.start)
                if getattr(rule, '_dtstart', None):
                    rule._dtstart = to_local(rule._dtstart, self.timezone)
                if getattr(rule, '_until', None):
                    rule._until = to_local(rule._until, self.timezone)
                rule = rrulestr(str(rule))
            except ValueError:
                # This might happen if only RDATEs and EXDATEs are present
                rule = rrulestr(self.recurrence, dtstart=start_local)

            # Make sure, the RDATEs and EXDATEs contain the start times
            for attribute in ('_exdate', '_rdate'):
                if hasattr(rule, attribute):
                    setattr(rule, attribute, [
                        to_local(date_, self.timezone).replace(
                            hour=start_local.hour, minute=start_local.minute
                        )
                        for date_ in getattr(rule, attribute)
                    ])

            # Generate the occurences and convert to UTC
            dates = [standardize_date(date_, self.timezone) for date_ in rule]

            # Make sure the start date is port of the reucrrence
            if self.start not in dates:
                dates.append(self.start)
                dates.sort()

        if localize:
            dates = [to_timezone(date_, self.timezone) for date_ in dates]

        if limit:
            max_year = datetime.today().year + 1
            dates = [date_ for date_ in dates if date_.year <= max_year]

        return dates

    def _update_occurrences(self):
        """ Updates the occurrences.

        Removes all occurrences if the event is not published or no start and
        end date/time is set. Only occurrences for this and next year are
        created.
        """

        # clear old occurrences
        self.occurrences = []

        # do not create occurrences unless the event is published
        if not self.state == 'published':
            return

        # do not create occurrences unless start and end is set
        if not self.start or not self.end:
            return

        # create all occurrences for this and next year
        for start in self.occurrence_dates():
            end = start + (self.end - self.start)
            name = '{0}-{1}'.format(self.name, start.date().isoformat())
            self.occurrences.append(
                Occurrence(
                    title=self.title,
                    name=name,
                    location=self.location,
                    tags=self.tags,
                    start=start,
                    end=end,
                    timezone=self.timezone
                )
            )

    def submit(self):
        """ Submit the event. """

        assert self.state == 'initiated'
        self.state = 'submitted'

    def publish(self):
        """ Publish the event.

        Publishing the event will generate the occurrences.
        """

        assert self.state == 'submitted' or self.state == 'withdrawn'
        self.state = 'published'

    def withdraw(self):
        """ Withdraw the event.

        Withdraw the event will delete the occurrences."""

        assert self.state == 'published'
        self.state = 'withdrawn'

    def as_ical(self, url=None):
        """ Returns the event and all its occurrences as iCalendar string. """

        modified = self.modified or self.created or datetime.utcnow()

        vevent = vEvent()
        vevent.add('uid', f'{self.name}@onegov.event')
        vevent.add('summary', self.title)
        vevent.add('dtstart', to_timezone(self.start, UTC))
        vevent.add('dtend', to_timezone(self.end, UTC))
        vevent.add('last-modified', modified)
        vevent.add('dtstamp', modified)
        vevent.add('location', self.location)
        vevent.add('description', self.description)
        vevent.add('categories', self.tags)
        if self.recurrence:
            vevent.add('rrule', self.icalendar_recurrence)
        if url:
            vevent.add('url', url)
        if self.coordinates:
            vevent.add('geo', (self.coordinates.lat, self.coordinates.lon))

        vcalendar = vCalendar()
        vcalendar.add('prodid', '-//OneGov//onegov.event//')
        vcalendar.add('version', '2.0')
        vcalendar.add_component(vevent)
        return vcalendar.to_ical()



from datetime import date
from datetime import datetime
from datetime import timedelta
from freezegun import freeze_time
from onegov.event import Event
from onegov.event import EventCollection
from onegov.event import Occurrence
from onegov.event import OccurrenceCollection
from onegov.gis import Coordinates
from sedate import replace_timezone
from sedate import standardize_date
import transaction


class DummyRequest(object):

    def link(self, item):
        return 'https://example.org/{}/{}'.format(
            item.__class__.__name__.lower(),
            getattr(item, 'name' or '?')
        )


def tzdatetime(year, month, day, hour, minute, timezone):
    return replace_timezone(datetime(year, month, day, hour, minute), timezone)


def test_event_collection(session):
    events = EventCollection(session)
    assert events.query().count() == 0

    event_1 = events.add(
        title='Squirrel Park Visit',
        start=datetime(2015, 6, 16, 9, 30),
        end=datetime(2015, 6, 16, 18, 00),
        timezone='US/Eastern',
        content={
            'description': '<em>Furry</em> things will happen!'
        },
        location='Squirrel Park',
        tags=['fun', 'animals'],
        recurrence=(
            'RRULE:FREQ=WEEKLY;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU;'
            'UNTIL=20150616T220000Z'
        )
    )
    event_1.submit()
    event_1.publish()
    event_2 = events.add(
        title='History of the Squirrel Park',
        start=datetime(2015, 6, 18, 14, 00),
        end=datetime(2015, 6, 18, 16, 00),
        timezone='US/Eastern',
        content={
            'description': 'Learn how the Park got so <em>furry</em>!'
        },
        location='Squirrel Park',
        tags=['history']
    )
    event_2.submit()
    event_2.publish()
    assert events.query().count() == 2

    assert events.by_id(event_1.id) == event_1
    assert events.by_id(event_2.id) == event_2

    events.delete(event_1)
    events.delete(event_2)
    assert events.query().count() == 0
    assert session.query(Occurrence).count() == 0


def test_event_collection_remove_stale_events(session):
    timezone = 'UTC'

    events = EventCollection(session)
    events.add(
        title='Event', timezone=timezone,
        start=datetime(2010, 6, 1, 10),
        end=datetime(2010, 6, 1, 10) + timedelta(hours=1),
    )
    assert events.query().count() == 1

    max_stale = standardize_date(datetime.today() + timedelta(days=2), 'UTC')
    events.remove_stale_events(max_stale=max_stale)
    assert events.query().count() == 0


def test_event_collection_pagination(session):
    events = EventCollection(session)

    assert events.page_index == 0
    assert events.pages_count == 0
    assert events.batch == []

    for year in range(2008, 2011):
        for month in range(1, 13):
            event = events.add(
                title='Event {0}-{1}'.format(year, month),
                start=datetime(year, month, 18, 14, 00),
                end=datetime(year, month, 18, 16, 00),
                timezone='US/Eastern',
                tags=['month-{0}'.format(month)],
                recurrence=(
                    f'RRULE:FREQ=WEEKLY;'
                    f'BYDAY=MO,TU,WE,TH,FR,SA,SU;'
                    f'UNTIL={year}{month:02}18T220000Z'
                )
            )
            event.submit()
            if year > 2008:
                event.publish()
            if year > 2009:
                event.withdraw()
    assert events.query().count() == 3 * 12

    events = EventCollection(session, state='initiated')
    assert events.subset_count == 0

    events = events.for_state('submitted')
    assert events.subset_count == 12
    assert all([e.start.year == 2008 for e in events.batch])
    assert all([e.start.month < 11 for e in events.batch])
    assert len(events.next.batch) == 12 - events.batch_size
    assert all([e.start.year == 2008 for e in events.next.batch])
    assert all([e.start.month > 10 for e in events.next.batch])
    assert events.page_by_index(1) == events.next

    events = events.for_state('published')
    assert events.subset_count == 12
    assert all([e.start.year == 2009 for e in events.batch])
    assert all([e.start.month < 11 for e in events.batch])
    assert len(events.next.batch) == 12 - events.batch_size
    assert all([e.start.year == 2009 for e in events.next.batch])
    assert all([e.start.month > 10 for e in events.next.batch])

    events = events.for_state('withdrawn')
    assert events.subset_count == 12
    assert all([e.start.year == 2010 for e in events.batch])
    assert all([e.start.month < 11 for e in events.batch])
    assert len(events.next.batch) == 12 - events.batch_size
    assert all([e.start.year == 2010 for e in events.next.batch])
    assert all([e.start.month > 10 for e in events.next.batch])


def test_occurrence_collection(session):
    event = EventCollection(session).add(
        title='Squirrel Park Visit',
        start=datetime(2015, 6, 16, 9, 30),
        end=datetime(2015, 6, 16, 18, 00),
        timezone='Pacific/Auckland',
        content={
            'description': '<em>Furry</em> things will happen!'
        },
        location='Squirrel Park',
        tags=['fun', 'park', 'animals'],
        recurrence=(
            'RRULE:FREQ=WEEKLY;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU;'
            'UNTIL=20150619T220000Z'
        )
    )
    event.submit()
    event.publish()
    event = EventCollection(session).add(
        title='History of the Squirrel Park',
        start=datetime(2015, 6, 18, 14, 00),
        end=datetime(2015, 6, 18, 16, 00),
        timezone='Europe/Zurich',
        content={
            'description': 'Learn how the Park got so <em>furry</em>!'
        },
        location='Squirrel Park',
        tags=['history']
    )
    event.submit()
    event.publish()

    occurrences = OccurrenceCollection(session)

    timezones = occurrences.used_timezones
    assert sorted(timezones) == ['Europe/Zurich', 'Pacific/Auckland']

    tags = occurrences.used_tags
    assert sorted(tags) == ['animals', 'fun', 'history', 'park']

    def query(**kwargs):
        return occurrences.for_filter(**kwargs).query(outdated=True)

    assert query().count() == 6

    assert query(tags=['animals']).count() == 5
    assert query(tags=['park']).count() == 5
    assert query(tags=['park', 'fun']).count() == 5
    assert query(tags=['history']).count() == 1
    assert query(tags=['history', 'fun']).count() == 6
    assert query(tags=[]).count() == 6

    assert query(start=date(2015, 6, 17)).count() == 5
    assert query(start=date(2015, 6, 18)).count() == 4
    assert query(start=date(2015, 6, 19)).count() == 2

    assert query(end=date(2015, 6, 19)).count() == 5
    assert query(end=date(2015, 6, 18)).count() == 4
    assert query(end=date(2015, 6, 17)).count() == 2

    assert query(start=date(2015, 6, 17), end=date(2015, 6, 18)).count() == 3
    assert query(start=date(2015, 6, 18), end=date(2015, 6, 18)).count() == 2

    assert query(start=date(2015, 6, 18), end=date(2015, 6, 18),
                 tags=['history']).count() == 1


def test_occurrence_collection_pagination(session):

    occurrences = OccurrenceCollection(session)
    assert occurrences.page_index == 0
    assert occurrences.pages_count == 0
    assert occurrences.batch == []

    for year in range(2008, 2011):
        for month in range(1, 13):
            event = EventCollection(session).add(
                title='Event {0}-{1}'.format(year, month),
                start=datetime(year, month, 18, 14, 00),
                end=datetime(year, month, 18, 16, 00),
                timezone='US/Eastern',
                tags=['month-{0}'.format(month)],
                recurrence=(
                    f'RRULE:FREQ=WEEKLY;'
                    f'BYDAY=MO,TU,WE,TH,FR,SA,SU;'
                    f'UNTIL={year}{month:02}21T220000Z'
                )
            )
            event.submit()
            if year > 2008:
                event.publish()
            if year > 2009:
                event.withdraw()

    class PatchedCollection(OccurrenceCollection):
        def query(self, **kwargs):
            return super().query(outdated=True, **kwargs)

    occurrences = PatchedCollection(session)
    assert occurrences.query().count() == 4 * 12

    occurrences = PatchedCollection(session)
    assert occurrences.subset_count == 48
    assert all([o.start.year == 2009 for o in occurrences.batch])
    assert occurrences.page_by_index(1) == occurrences.next

    occurrences = PatchedCollection(session, start=date(2009, 12, 1))
    assert occurrences.subset_count == 4
    assert all([o.start.year == 2009 and o.start.month == 12
                for o in occurrences.batch])

    occurrences = PatchedCollection(session, end=date(2009, 1, 31))
    assert occurrences.subset_count == 4
    assert all([o.start.year == 2009 and o.start.month == 1
                for o in occurrences.batch])

    occurrences = PatchedCollection(session,
                                    start=date(2009, 5, 1),
                                    end=date(2009, 6, 30))
    assert occurrences.subset_count == 8
    assert all([o.start.year == 2009 and o.start.month in [5, 6]
                for o in occurrences.batch])

    occurrences = PatchedCollection(session, tags=['month-7', 'month-8'])

    assert occurrences.subset_count == 8
    assert all([o.start.year == 2009 and o.start.month in [7, 8]
                for o in occurrences.batch])

    occurrences = PatchedCollection(session,
                                    start=date(2009, 5, 1),
                                    end=date(2009, 6, 30),
                                    tags=['month-6'])
    assert occurrences.subset_count == 4
    assert all([o.start.year == 2009 and o.start.month == 6
                for o in occurrences.batch])

    occurrences = PatchedCollection(session).for_filter()
    assert occurrences.start is None
    assert occurrences.end is None
    assert occurrences.tags == []

    occurrences = PatchedCollection(session,
                                    start=date(2009, 5, 1),
                                    end=date(2009, 6, 30),
                                    tags=['month-6']).for_filter()
    assert occurrences.start == date(2009, 5, 1)
    assert occurrences.end == date(2009, 6, 30)
    assert occurrences.tags == ['month-6']

    occurrences = occurrences.for_filter(start=date(2010, 5, 1))
    assert occurrences.start == date(2010, 5, 1)
    assert occurrences.end == date(2009, 6, 30)
    assert occurrences.tags == ['month-6']

    occurrences = occurrences.for_filter(end=None)
    assert occurrences.start == date(2010, 5, 1)
    assert occurrences.end is None
    assert occurrences.tags == ['month-6']

    occurrences = occurrences.for_filter(tags=[])
    assert occurrences.start == date(2010, 5, 1)
    assert occurrences.end is None
    assert occurrences.tags == []

    occurrences = occurrences.for_filter(tags=['a', 'b'])
    assert occurrences.start == date(2010, 5, 1)
    assert occurrences.end is None
    assert occurrences.tags == ['a', 'b']

    occurrences = occurrences.for_filter(tag='c')
    assert occurrences.start == date(2010, 5, 1)
    assert occurrences.end is None
    assert occurrences.tags == ['a', 'b', 'c']

    occurrences = occurrences.for_filter(tag='a')
    assert occurrences.start == date(2010, 5, 1)
    assert occurrences.end is None
    assert occurrences.tags == ['b', 'c']


def test_occurrence_collection_outdated(session):
    today = date.today()
    for year in (today.year - 1, today.year, today.year + 1):
        event = EventCollection(session).add(
            title='Event {0}-{1}'.format(year, today.month),
            start=datetime(year, today.month, today.day, 0, 0),
            end=datetime(year, today.month, today.day, 23, 59),
            timezone='US/Eastern'
        )
        event.submit()
        event.publish()

    occurrences = OccurrenceCollection(session)
    assert occurrences.query().count() == 2
    assert occurrences.query(outdated=True).count() == 3

    occurrences = occurrences.for_filter(start=date(today.year - 1, 1, 1))
    assert occurrences.query().count() == 3


def test_unique_names(session):
    events = EventCollection(session)
    added = [
        events.add(
            title='Squirrel Park Visit',
            start=datetime(2015, 6, 16, 9, 30),
            end=datetime(2015, 6, 16, 18, 00),
            timezone='US/Eastern'
        ) for x in range(11)
    ]
    assert added[0].name == 'squirrel-park-visit'
    assert added[1].name == 'squirrel-park-visit-1'
    assert added[2].name == 'squirrel-park-visit-2'
    assert added[3].name == 'squirrel-park-visit-3'
    assert added[4].name == 'squirrel-park-visit-4'
    assert added[5].name == 'squirrel-park-visit-5'
    assert added[6].name == 'squirrel-park-visit-6'
    assert added[7].name == 'squirrel-park-visit-7'
    assert added[8].name == 'squirrel-park-visit-8'
    assert added[9].name == 'squirrel-park-visit-9'
    assert added[10].name == 'squirrel-park-visit-10'

    events.delete(added[6])
    event = events.add(
        title='Squirrel Park Visit',
        start=datetime(2015, 6, 16, 9, 30),
        end=datetime(2015, 6, 16, 18, 00),
        timezone='US/Eastern',
        recurrence=(
            'RRULE:FREQ=WEEKLY;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU;'
            'UNTIL=20150620T220000Z'
        )
    )
    assert event.name == 'squirrel-park-visit-6'

    event.submit()
    event.publish()
    assert event.occurrences[0].name == 'squirrel-park-visit-6-2015-06-16'
    assert event.occurrences[1].name == 'squirrel-park-visit-6-2015-06-17'
    assert event.occurrences[2].name == 'squirrel-park-visit-6-2015-06-18'
    assert event.occurrences[3].name == 'squirrel-park-visit-6-2015-06-19'
    assert event.occurrences[4].name == 'squirrel-park-visit-6-2015-06-20'

    assert events.by_name('test') is None
    assert events.by_name('squirrel-park-visit-6') == event

    occurrences = OccurrenceCollection(session)
    assert occurrences.by_name('test') is None

    occurrence = occurrences.by_name('squirrel-park-visit-6-2015-06-20')
    assert occurrence == event.occurrences[4]


def test_unicode(session):
    event = EventCollection(session).add(
        title='Salon du mieux-vivre, 16e édition',
        start=datetime(2015, 6, 16, 9, 30),
        end=datetime(2015, 6, 16, 18, 00),
        timezone='Europe/Zurich',
        content={
            'description': 'Rendez-vous automnal des médecines.'
        },
        location='Salon du mieux-vivre à Saignelégier',
        tags=['salons', 'congrès']
    )
    event.submit()
    event.publish()
    event = EventCollection(session).add(
        title='Témoins de Jéhovah',
        start=datetime(2015, 6, 18, 14, 00),
        end=datetime(2015, 6, 18, 16, 00),
        timezone='Europe/Zurich',
        content={
            'description': 'Congrès en français et espagnol.'
        },
        location='Salon du mieux-vivre à Saignelégier',
        tags=['témoins']
    )
    event.submit()
    event.publish()
    session.flush()

    occurrences = OccurrenceCollection(session)

    assert sorted(occurrences.used_tags) == ['congrès', 'salons', 'témoins']

    assert occurrences.query(outdated=True).count() == 2

    occurrences = occurrences.for_filter(tags=['congrès'])
    occurrence = occurrences.query(outdated=True).one()
    assert occurrence.title == 'Salon du mieux-vivre, 16e édition'
    assert occurrence.location == 'Salon du mieux-vivre à Saignelégier'
    assert sorted(occurrence.tags) == ['congrès', 'salons']
    assert occurrence.event.description \
        == 'Rendez-vous automnal des médecines.'

    occurrences = occurrences.for_filter(tags=['témoins'])
    occurrence = occurrences.query(outdated=True).one()
    assert occurrence.title == 'Témoins de Jéhovah'
    assert occurrence.location == 'Salon du mieux-vivre à Saignelégier'
    assert occurrence.tags == ['témoins']
    assert occurrence.event.description == 'Congrès en français et espagnol.'


def test_as_ical(session):
    occurrences = OccurrenceCollection(session)
    ical = occurrences.as_ical(DummyRequest()).decode().strip().split('\r\n')
    assert sorted(ical) == sorted([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'END:VCALENDAR',
    ])

    events = EventCollection(session)
    with freeze_time("2014-01-01"):
        event = events.add(
            title='Squirrel Park Visit',
            start=datetime(2015, 6, 16, 9, 30),
            end=datetime(2015, 6, 16, 18, 00),
            timezone='US/Eastern',
            content={
                'description': '<em>Furry</em> things will happen!'
            },
            location='Squirrel Park',
            tags=['fun', 'animals'],
            recurrence=(
                'RRULE:FREQ=WEEKLY;'
                'BYDAY=MO,TU,WE,TH,FR,SA,SU;'
                'UNTIL=20150616T220000Z'
            ),
            coordinates=Coordinates(47.051752750515746, 8.305739625357093)
        )
        event.submit()
        event.publish()

        event = events.add(
            title='History of the Squirrel Park',
            start=datetime(2015, 6, 18, 14, 00),
            end=datetime(2015, 6, 18, 16, 00),
            timezone='US/Eastern',
            content={
                'description': 'Learn how the Park got so <em>furry</em>!'
            },
            location='Squirrel Park',
            tags=['history'],
            coordinates=Coordinates(47.051752750515746, 8.305739625357093)
        )
        event.submit()
        event.publish()

        session.flush()

    occurrences = OccurrenceCollection(session)
    ical = occurrences.as_ical(DummyRequest()).decode().strip().split('\r\n')
    occurrences.query().all()
    assert sorted(ical) == sorted([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'END:VCALENDAR',
    ])

    occurrences = occurrences.for_filter(start=date(2015, 6, 1))
    ical = occurrences.as_ical(DummyRequest()).decode().strip().split('\r\n')
    assert sorted(ical) == sorted([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'SUMMARY:Squirrel Park Visit',
        'UID:squirrel-park-visit@onegov.event',
        'DTSTART;VALUE=DATE-TIME:20150616T133000Z',
        'DTEND;VALUE=DATE-TIME:20150616T220000Z',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        'RRULE:FREQ=WEEKLY;UNTIL=20150616T220000Z;BYDAY=MO,TU,WE,TH,FR,SA,SU',
        'DESCRIPTION:<em>Furry</em> things will happen!',
        'CATEGORIES:fun,animals',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/event/squirrel-park-visit',
        'END:VEVENT',
        'BEGIN:VEVENT',
        'UID:history-of-the-squirrel-park@onegov.event',
        'SUMMARY:History of the Squirrel Park',
        'DTSTART;VALUE=DATE-TIME:20150618T180000Z',
        'DTEND;VALUE=DATE-TIME:20150618T200000Z',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        'DESCRIPTION:Learn how the Park got so <em>furry</em>!',
        'CATEGORIES:history',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/event/history-of-the-squirrel-park',
        'END:VEVENT',
        'END:VCALENDAR'
    ])

    occurrences = occurrences.for_filter(
        start=date(2015, 6, 18),
        end=date(2018, 6, 18),
        tags=['history'])
    ical = occurrences.as_ical(DummyRequest()).decode().strip().split('\r\n')
    assert sorted(ical) == sorted([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'UID:history-of-the-squirrel-park@onegov.event',
        'SUMMARY:History of the Squirrel Park',
        'DTSTART;VALUE=DATE-TIME:20150618T180000Z',
        'DTEND;VALUE=DATE-TIME:20150618T200000Z',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        'DESCRIPTION:Learn how the Park got so <em>furry</em>!',
        'CATEGORIES:history',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/event/history-of-the-squirrel-park',
        'END:VEVENT',
        'END:VCALENDAR'
    ])


def test_from_import(session):
    events = EventCollection(session)

    assert events.from_import([
        Event(
            state='initiated',
            title='Title A',
            location='Location A',
            tags=['Tag A.1', 'Tag A.2'],
            start=tzdatetime(2015, 6, 16, 9, 30, 'US/Eastern'),
            end=tzdatetime(2015, 6, 16, 18, 00, 'US/Eastern'),
            timezone='US/Eastern',
            description='Description A',
            organizer='Organizer A',
            recurrence=(
                'RRULE:FREQ=WEEKLY;'
                'UNTIL=20150616T220000Z;'
                'BYDAY=MO,TU,WE,TH,FR,SA,SU'
            ),
            coordinates=Coordinates(48.051752750515746, 9.305739625357093),
            source='import-1-A'
        ),
        Event(
            state='initiated',
            title='Title B',
            location='Location B',
            tags=['Tag B.1', 'Tag B.2'],
            start=tzdatetime(2015, 6, 16, 9, 30, 'US/Eastern'),
            end=tzdatetime(2015, 6, 16, 18, 00, 'US/Eastern'),
            timezone='US/Eastern',
            description='Description B',
            organizer='Organizer B',
            recurrence=(
                'RRULE:FREQ=WEEKLY;'
                'UNTIL=20150616T220000Z;'
                'BYDAY=MO,TU,WE,TH,FR,SA,SU'
            ),
            coordinates=Coordinates(48.051752750515746, 9.305739625357093),
            source='import-1-B'
        ),
    ]) == (2, 0, 0)

    assert events.from_import([
        Event(
            state='initiated',
            title='Title C',
            location='Location C',
            tags=['Tag C.1', 'Tag C.2'],
            start=tzdatetime(2015, 6, 16, 9, 30, 'US/Eastern'),
            end=tzdatetime(2015, 6, 16, 18, 00, 'US/Eastern'),
            timezone='US/Eastern',
            description='Description C',
            organizer='Organizer C',
            recurrence=(
                'RRULE:FREQ=WEEKLY;'
                'UNTIL=20150616T220000Z;'
                'BYDAY=MO,TU,WE,TH,FR,SA,SU'
            ),
            coordinates=Coordinates(48.051752750515746, 9.305739625357093),
            source='import-2-C'
        )
    ]) == (1, 0, 0)

    # Already imported
    assert events.from_import([
        Event(
            state='initiated',
            title='Title C',
            location='Location C',
            tags=['Tag C.1', 'Tag C.2'],
            start=tzdatetime(2015, 6, 16, 9, 30, 'US/Eastern'),
            end=tzdatetime(2015, 6, 16, 18, 00, 'US/Eastern'),
            timezone='US/Eastern',
            description='Description C',
            organizer='Organizer C',
            recurrence=(
                'RRULE:FREQ=WEEKLY;'
                'UNTIL=20150616T220000Z;'
                'BYDAY=MO,TU,WE,TH,FR,SA,SU'
            ),
            coordinates=Coordinates(48.051752750515746, 9.305739625357093),
            source='import-2-C'
        )
    ]) == (0, 0, 0)

    # Update and purge
    assert events.from_import([
        Event(
            state='initiated',
            title='Title',
            location='Location A',
            tags=['Tag A.1', 'Tag A.2'],
            start=tzdatetime(2015, 6, 16, 9, 30, 'US/Eastern'),
            end=tzdatetime(2015, 6, 16, 18, 00, 'US/Eastern'),
            timezone='US/Eastern',
            description='Description A',
            organizer='Organizer A',
            recurrence=(
                'RRULE:FREQ=WEEKLY;'
                'UNTIL=20150616T220000Z;'
                'BYDAY=MO,TU,WE,TH,FR,SA,SU'
            ),
            coordinates=Coordinates(48.051752750515746, 9.305739625357093),
            source='import-1-A'
        ),
    ], 'import-1') == (0, 1, 1)
    assert events.subset_count == 2

    # Withdraw
    events.by_name('title-c').withdraw()
    assert events.from_import([
        Event(
            state='initiated',
            title='Title C',
            location='Location C',
            tags=['Tag C.1', 'Tag C.2'],
            start=tzdatetime(2015, 6, 16, 9, 30, 'US/Eastern'),
            end=tzdatetime(2015, 6, 16, 18, 00, 'US/Eastern'),
            timezone='US/Eastern',
            description='Description C',
            organizer='Organizer C',
            recurrence=(
                'RRULE:FREQ=WEEKLY;'
                'UNTIL=20150616T220000Z;'
                'BYDAY=MO,TU,WE,TH,FR,SA,SU'
            ),
            coordinates=Coordinates(48.051752750515746, 9.305739625357093),
            source='import-2-C'
        )
    ]) == (0, 0, 0)
    assert events.by_name('title-c').state == 'withdrawn'


def test_from_ical(session):
    events = EventCollection(session)

    events.from_ical('\n'.join([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'END:VCALENDAR',
    ]))
    assert events.query().count() == 0

    # UTC-date
    events.from_ical('\n'.join([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'SUMMARY:Squirrel Park Virsit',
        'UID:squirrel-park-visit@onegov.event',
        'DTSTART;VALUE=DATE-TIME:20150616T133100Z',
        'DTEND;VALUE=DATE-TIME:20150616T220100Z',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        (
            'RRULE:FREQ=WEEKLY;'
            'UNTIL=20150619T220000Z;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        ),
        'DESCRIPTION:<em>Furri</em> things will happen!',
        'CATEGORIES:fun,animals',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Par',
        'GEO:48.051752750515746;9.305739625357093',
        'URL:https://example.org/event/squirrel-park-visit',
        'END:VEVENT',
        'END:VCALENDAR'
    ]))
    transaction.commit()
    event = events.query().one()
    assert event.title == 'Squirrel Park Virsit'
    assert event.description == '<em>Furri</em> things will happen!'
    assert event.location == 'Squirrel Par'
    assert event.start == tzdatetime(2015, 6, 16, 9, 31, 'US/Eastern')
    assert str(event.start.tzinfo) == 'UTC'
    assert event.end == tzdatetime(2015, 6, 16, 18, 1, 'US/Eastern')
    assert str(event.end.tzinfo) == 'UTC'
    assert event.timezone == 'Europe/Zurich'
    assert event.recurrence == (
        'RRULE:FREQ=WEEKLY;'
        'UNTIL=20150619T220000Z;'
        'BYDAY=MO,TU,WE,TH,FR,SA,SU'
    )
    assert [o.start.day for o in event.occurrences] == [16, 17, 18, 19]
    assert sorted(event.tags) == ['animals', 'fun']
    assert int(event.coordinates.lat) == 48
    assert int(event.coordinates.lon) == 9

    # update
    events.from_ical('\n'.join([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'SUMMARY:Squirrel Park Visit',
        'UID:squirrel-park-visit@onegov.event',
        'DTSTART;VALUE=DATE-TIME:20150616T133000Z',
        'DTEND;VALUE=DATE-TIME:20150616T220000Z',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        (
            'RRULE:FREQ=WEEKLY;'
            'UNTIL=20150620T220000Z;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        ),
        'DESCRIPTION:<em>Furry</em> things will happen!',
        'CATEGORIES:fun,animals',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/event/squirrel-park-visit',
        'END:VEVENT',
        'END:VCALENDAR'
    ]))
    transaction.commit()
    event = events.query().one()
    assert event.title == 'Squirrel Park Visit'
    assert event.description == '<em>Furry</em> things will happen!'
    assert event.location == 'Squirrel Park'
    assert event.start == tzdatetime(2015, 6, 16, 9, 30, 'US/Eastern')
    assert str(event.start.tzinfo) == 'UTC'
    assert event.end == tzdatetime(2015, 6, 16, 18, 0, 'US/Eastern')
    assert str(event.end.tzinfo) == 'UTC'
    assert event.timezone == 'Europe/Zurich'
    assert event.recurrence == (
        'RRULE:FREQ=WEEKLY;'
        'UNTIL=20150620T220000Z;'
        'BYDAY=MO,TU,WE,TH,FR,SA,SU'
    )
    assert [o.start.day for o in event.occurrences] == [16, 17, 18, 19, 20]
    assert sorted(event.tags) == ['animals', 'fun']
    assert int(event.coordinates.lat) == 47
    assert int(event.coordinates.lon) == 8

    # date
    events.from_ical('\n'.join([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'SUMMARY:Squirrel Park Virsit',
        'UID:squirrel-park-visit@onegov.event',
        'DTSTART;VALUE=DATE:20150616',
        'DTEND;VALUE=DATE:20150616',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        (
            'RRULE:FREQ=WEEKLY;'
            'UNTIL=20150616T220000Z;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        ),
        'DESCRIPTION:<em>Furri</em> things will happen!',
        'CATEGORIES:fun,animals',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Par',
        'GEO:48.051752750515746;9.305739625357093',
        'URL:https://example.org/event/squirrel-park-visit',
        'END:VEVENT',
        'END:VCALENDAR'
    ]))
    transaction.commit()
    event = events.query().one()
    assert event.start == tzdatetime(2015, 6, 16, 0, 0, 'Europe/Zurich')
    assert event.end == tzdatetime(2015, 6, 16, 23, 59, 'Europe/Zurich')

    # relative date-time
    events.from_ical('\n'.join([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'SUMMARY:Squirrel Park Visit',
        'UID:squirrel-park-visit@onegov.event',
        'DTSTART;VALUE=DATE-TIME:20150616T133000',
        'DTEND;VALUE=DATE-TIME:20150616T220000',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        (
            'RRULE:FREQ=WEEKLY;'
            'UNTIL=20150616T220000Z;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        ),
        'DESCRIPTION:<em>Furry</em> things will happen!',
        'CATEGORIES:fun,animals',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/event/squirrel-park-visit',
        'END:VEVENT',
        'END:VCALENDAR'
    ]))
    transaction.commit()
    event = events.query().one()
    assert event.start == tzdatetime(2015, 6, 16, 13, 30, 'Europe/Zurich')
    assert event.end == tzdatetime(2015, 6, 16, 22, 0, 'Europe/Zurich')

    # start and duration
    events.from_ical('\n'.join([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'SUMMARY:Squirrel Park Visit',
        'UID:squirrel-park-visit@onegov.event',
        'DTSTART;VALUE=DATE-TIME:20150616T133000',
        'DURATION:PT8H30M',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        (
            'RRULE:FREQ=WEEKLY;'
            'UNTIL=20150616T220000Z;'
            'BYDAY=MO,TU,WE,TH,FR,SA,SU'
        ),
        'DESCRIPTION:<em>Furry</em> things will happen!',
        'CATEGORIES:fun,animals',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirrel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/event/squirrel-park-visit',
        'END:VEVENT',
        'END:VCALENDAR'
    ]))
    transaction.commit()
    event = events.query().one()
    assert event.start == tzdatetime(2015, 6, 16, 13, 30, 'Europe/Zurich')
    assert event.end == tzdatetime(2015, 6, 16, 22, 0, 'Europe/Zurich')



from datetime import datetime
from lxml import etree
from onegov.core.utils import module_path
from onegov.event.utils import GuidleExportData
from pytest import mark
from sedate import replace_timezone


def tzdatetime(year, month, day, hour, minute, seconds=0, microseconds=0):
    return replace_timezone(
        datetime(year, month, day, hour, minute, seconds, microseconds),
        timezone='Europe/Zurich'
    )


@mark.parametrize("xml", [
    module_path('onegov.event', 'tests/fixtures/guidle.xml'),
])
def test_import_guidle(session, xml):
    offers = list(GuidleExportData(etree.parse(xml)).offers())
    assert len(offers) == 1

    assert offers[0].uid == '551262854'
    assert offers[0].title == "Theatervorstellung"
    assert offers[0].description == (
        "Lorem ipsum\n\n"
        "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam "
        "nonumy eirmod tempor invidunt ut labore et dolore magna "
        "aliquyam.\r\n\r\n"
        "anschliessend Apéro\r\n\r\n"
        "1-Tages- oder 2-Tagespass erhältlich\r\n\r\n"
        "www.theatermenzingen.ch"
    )
    assert offers[0].organizer == (
        "Theatervereinigung Menzingen, Peter Muster, "
        "peter.muster@theatermenzingen.ch, +41 41 755 35 46"
    )
    assert offers[0].location == (
        "Zentrum Schützenmatt, Luegetenstrasse 3, 6313 Menzingen"
    )
    assert int(offers[0].coordinates.lat) == 47
    assert int(offers[0].coordinates.lon) == 8
    assert offers[0].tags() == (
        {'Konzert Pop / Rock / Jazz', 'Kulinarik'},
        set()
    )
    assert offers[0].tags({'Konzert Pop / Rock / Jazz': 'Konzert'}) == (
        {'Konzert'}, {'Kulinarik'}
    )

    schedules = list(offers[0].schedules())
    assert len(schedules) == 4

    assert schedules[0].start == tzdatetime(2018, 10, 26, 20, 0)
    assert schedules[0].end == tzdatetime(2018, 10, 26, 22, 30)
    assert schedules[0].recurrence == (
        'RRULE:FREQ=WEEKLY;BYDAY=TU,FR,SA;UNTIL=20181104T0000Z'
    )

    assert schedules[1].start == tzdatetime(2018, 10, 28, 17, 0)
    assert schedules[1].end == tzdatetime(2018, 10, 28, 23, 59, 59, 999999)
    assert schedules[1].recurrence == ''

    assert schedules[2].start == tzdatetime(2018, 6, 15, 8, 0)
    assert schedules[2].end == tzdatetime(2018, 6, 15, 23, 59, 59, 999999)
    assert schedules[2].recurrence == (
        'RRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR,SA,SU;UNTIL=20181016T0000Z'
    )

    assert schedules[3].start == tzdatetime(2018, 8, 18, 0, 0)
    assert schedules[3].end == tzdatetime(2018, 8, 18, 23, 59, 59, 999999)
    assert schedules[3].recurrence == (
        'RRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR,SA,SU;UNTIL=20181101T0000Z'
    )



from datetime import date
from datetime import datetime
from datetime import timedelta
from icalendar import Calendar as vCalendar
from onegov.core.collection import Pagination
from onegov.core.utils import increment_name
from onegov.core.utils import normalize_for_url
from onegov.event.models import Event
from onegov.gis import Coordinates
from pytz import UTC
from sedate import as_datetime
from sedate import replace_timezone
from sedate import standardize_date
from sedate import to_timezone
from sqlalchemy import and_
from sqlalchemy import or_


class EventCollection(Pagination):

    """ Manage a list of events. """

    def __init__(self, session, page=0, state=None):
        self.session = session
        self.page = page
        self.state = state

    def __eq__(self, other):
        return self.state == other.state and self.page == other.page

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(self.session, index, self.state)

    def for_state(self, state):
        """ Returns a new instance of the collection with the given state. """

        return self.__class__(self.session, 0, state)

    def query(self):
        query = self.session.query(Event)
        if self.state:
            query = query.filter(Event.state == self.state)
        query = query.order_by(Event.start)
        return query

    def _get_unique_name(self, name):
        """ Create a unique, URL-friendly name. """

        # it's possible for `normalize_for_url` to return an empty string...
        name = normalize_for_url(name) or "event"

        session = self.session
        while session.query(Event.name).filter(Event.name == name).first():
            name = increment_name(name)

        return name

    def add(self, title, start, end, timezone, autoclean=True, **optional):
        """ Add a new event.

        A unique, URL-friendly name is created automatically for this event
        using the title and optionally numbers for duplicate names.

        Every time a new event is added, old, stale events are deleted unless
        disabled with the ``autoclean`` option.

        Returns the created event or None if already automatically deleted.
        """
        if not start.tzinfo:
            start = replace_timezone(start, timezone)
        if not end.tzinfo:
            end = replace_timezone(end, timezone)

        event = Event(
            state='initiated',
            title=title,
            start=start,
            end=end,
            timezone=timezone,
            name=self._get_unique_name(title),
            **optional
        )

        self.session.add(event)
        self.session.flush()

        if autoclean:
            self.remove_stale_events()

        return event

    def delete(self, event):
        """ Delete an event. """

        self.session.delete(event)
        self.session.flush()

    def remove_stale_events(self, max_stale=None):
        """ Remove events which have never been submitted and are created more
        than five days ago.

        """

        if max_stale is None:
            max_stale = datetime.utcnow() - timedelta(days=5)
            max_stale = standardize_date(max_stale, 'UTC')

        events = self.session.query(Event).filter(
            Event.state == 'initiated',
            and_(
                or_(Event.created < max_stale, Event.created.is_(None)),
                or_(Event.modified < max_stale, Event.modified.is_(None))
            )
        )
        for event in events:
            self.session.delete(event)

        self.session.flush()

    def by_name(self, name):
        """ Returns an event by its URL-friendly name. """

        query = self.session.query(Event).filter(Event.name == name)
        return query.first()

    def by_id(self, id):
        """ Return an event by its id. """

        query = self.session.query(Event).filter(Event.id == id)
        return query.first()

    def from_import(self, events, purge=None):
        """ Add or updates the given events.

        Only updates events which have changed. Doesn't change the states of
        events allowing to permanently withdraw imported events.

        Optionally removes all events with the given meta-source-prefix not
        present in the given events.

        """

        sources = set((event.source for event in events))
        assert all(sources) and (len(sources) == len(events))

        if purge:
            assert all((s.startswith(purge) for s in sources))
            query = self.session.query(Event.meta['source'].label('source'))
            query = query.filter(Event.meta['source'].astext.startswith(purge))
            purge = set((r.source for r in query)) - sources

        added = 0
        updated = 0

        for event in events:
            existing = self.session.query(Event).filter(
                Event.meta['source'] == event.meta['source']
            ).first()

            if existing:
                if (
                    existing.title != event.title or
                    existing.location != event.location or
                    set(existing.tags) != set(event.tags) or
                    existing.timezone != event.timezone or
                    existing.start != event.start or
                    existing.end != event.end or
                    existing.content != event.content or
                    existing.coordinates != event.coordinates or
                    existing.recurrence != event.recurrence
                ):
                    updated += 1
                    state = existing.state  # avoid updating occurrences
                    existing.state = 'initiated'
                    existing.title = event.title
                    existing.location = event.location
                    existing.tags = event.tags
                    existing.timezone = event.timezone
                    existing.start = event.start
                    existing.end = event.end
                    existing.content = event.content
                    existing.coordinates = event.coordinates
                    existing.recurrence = event.recurrence
                    existing.state = state

            else:
                added += 1
                event.name = self._get_unique_name(event.title)
                event.state = 'initiated'
                self.session.add(event)
                event.submit()
                event.publish()

        if purge:
            query = self.session.query(Event)
            query = query.filter(Event.meta['source'].in_(purge))
            for event in query:
                event.state = 'withdrawn'  # remove occurrences
                self.session.delete(event)

        self.session.flush()

        return added, updated, len(purge) if purge else 0

    def from_ical(self, ical):
        """ Imports the events from an iCalender string.

        We assume the timezone to be Europe/Zurich!

        """
        events = []

        cal = vCalendar.from_ical(ical)
        for vevent in cal.walk('vevent'):
            timezone = 'Europe/Zurich'
            start = vevent.get('dtstart')
            start = start.dt if start else None
            if type(start) is date:
                start = replace_timezone(as_datetime(start), timezone)
            elif type(start) is datetime:
                if start.tzinfo is None:
                    start = standardize_date(start, timezone)
                else:
                    start = to_timezone(start, UTC)

            end = vevent.get('dtend')
            end = end.dt if end else None
            if type(end) is date:
                end = replace_timezone(as_datetime(end), timezone)
                end = end + timedelta(days=1, minutes=-1)
            elif type(end) is datetime:
                if end.tzinfo is None:
                    end = standardize_date(end, timezone)
                else:
                    end = to_timezone(end, UTC)

            duration = vevent.get('duration')
            duration = duration.dt if duration else None
            if start and not end and duration:
                end = start + duration

            if not start or not end:
                raise(ValueError("Invalid date"))

            recurrence = vevent.get('rrule', '')
            if recurrence:
                recurrence = 'RRULE:{}'.format(recurrence.to_ical().decode())

            coordinates = vevent.get('geo')
            if coordinates:
                coordinates = Coordinates(
                    coordinates.latitude, coordinates.longitude
                )

            tags = vevent.get('categories')
            if tags:
                # categories may be in lists or they may be single values
                # whose 'cats' member contains the texts
                if not hasattr(tags, '__iter__'):
                    tags = [tags]

                tags = [str(c) for tag in tags for c in tag.cats]

            uid = str(vevent.get('uid', ''))
            title = str(vevent.get('summary', ''))
            description = str(vevent.get('description', ''))
            organizer = str(vevent.get('organizer', ''))
            location = str(vevent.get('location', ''))

            events.append(
                Event(
                    state='initiated',
                    title=title,
                    start=start,
                    end=end,
                    timezone=timezone,
                    recurrence=recurrence,
                    description=description,
                    organizer=organizer,
                    location=location,
                    coordinates=coordinates,
                    tags=tags or [],
                    source=f'ical-{uid}',
                )
            )

        return self.from_import(events)



from dateutil.rrule import rrulestr
from sedate import standardize_date


def as_rdates(recurrence):
    return '\n'.join(
        f"RDATE:{d.strftime('%Y-%m-%dT%H%M%SZ')}" for d in (
            standardize_date(date, timezone='UTC')
            for date in rrulestr(recurrence)
        )
    )



from cached_property import cached_property
from datetime import date
from datetime import datetime
from datetime import timedelta
from icalendar import Calendar as vCalendar
from icalendar import Event as vEvent
from onegov.core.collection import Pagination
from onegov.core.utils import get_unique_hstore_keys
from onegov.event.models import Event
from onegov.event.models import Occurrence
from pytz import UTC
from sedate import as_datetime
from sedate import replace_timezone
from sedate import standardize_date
from sedate import to_timezone
from sqlalchemy import and_
from sqlalchemy import distinct
from sqlalchemy import or_
from sqlalchemy.dialects.postgresql import array


class OccurrenceCollection(Pagination):
    """ Manages a list of occurrences.

    Occurrences are read only (no ``add`` method here), they are generated
    automatically when adding a new event.

    Occurrences can be filtered by start and end dates as well as tags.

    By default, only current occurrences are returned.
    """

    def __init__(self, session, page=0, start=None, end=None, tags=None):
        self.session = session
        self.page = page
        self.start = start
        self.end = end
        self.tags = tags if tags else []

    def __eq__(self, other):
        return self.page == other.page

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(
            self.session, index, self.start, self.end, self.tags
        )

    def for_filter(self, **kwargs):
        """ Returns a new instance of the collection.

        Copies the current filters if not specified. Also adds or removes a
        single tag if given.
        """

        start = kwargs['start'] if 'start' in kwargs else self.start
        end = kwargs['end'] if 'end' in kwargs else self.end
        tags = kwargs['tags'] if 'tags' in kwargs else list(self.tags)

        if 'tag' in kwargs:
            tag = kwargs['tag']
            if tag in tags:
                tags.remove(tag)
            elif tag is not None:
                tags.append(tag)

        return self.__class__(self.session, 0, start, end, tags)

    @cached_property
    def used_timezones(self):
        """ Returns a list of all the timezones used by the occurrences. """

        return [
            tz[0] for tz in self.session.query(distinct(Occurrence.timezone))
        ]

    @cached_property
    def used_tags(self):
        """ Returns a list of all the tags used by the occurrences.

        This could be solve possibly more effienciently with the skey function
        currently not supported by SQLAlchemy (e.g.
        ``select distinct(skeys(tags))``), see
        http://stackoverflow.com/q/12015942/3690178

        """

        return get_unique_hstore_keys(self.session, Occurrence._tags)

    def query(self, outdated=False):
        """ Queries occurrences with the set parameters.

        Finds all occurrences with any of the set tags and within the set
        start and end date. Start and end date are assumed to be dates only and
        therefore without a timezone - we search for the given date in the
        timezone of the occurrence!.

        If no start date is set and ``outdated`` is not set, only current
        occurrences are returned.
        """

        query = self.session.query(Occurrence)

        if self.start is not None:
            assert type(self.start) is date
            start = as_datetime(self.start)

            expressions = []
            for timezone in self.used_timezones:
                localized_start = replace_timezone(start, timezone)
                localized_start = standardize_date(localized_start, timezone)
                expressions.append(
                    and_(
                        Occurrence.timezone == timezone,
                        Occurrence.start >= localized_start
                    )
                )

            query = query.filter(or_(*expressions))
        elif not outdated:
            start = as_datetime(date.today())

            expressions = []
            for timezone in self.used_timezones:
                localized_start = replace_timezone(start, timezone)
                localized_start = standardize_date(localized_start, timezone)
                expressions.append(
                    and_(
                        Occurrence.timezone == timezone,
                        Occurrence.start >= localized_start
                    )
                )

            query = query.filter(or_(*expressions))

        if self.end is not None:
            assert type(self.end) is date
            end = as_datetime(self.end)
            end = end + timedelta(days=1)

            expressions = []
            for timezone in self.used_timezones:
                localized_end = replace_timezone(end, timezone)
                localized_end = standardize_date(localized_end, timezone)
                expressions.append(
                    and_(
                        Occurrence.timezone == timezone,
                        Occurrence.end < localized_end
                    )
                )

            query = query.filter(or_(*expressions))

        if self.tags:
            query = query.filter(Occurrence._tags.has_any(array(self.tags)))

        query = query.order_by(Occurrence.start, Occurrence.title)

        return query

    def by_name(self, name):
        """ Returns an occurrence by its URL-friendly name.

        The URL-friendly name is automatically constructed as follows:

        ``unique name of the event``-``date of the occurrence``

        e.g.

        ``squirrel-park-visit-6-2015-06-20``

        """

        query = self.session.query(Occurrence).filter(Occurrence.name == name)
        return query.first()

    def as_ical(self, request):
        """ Returns the the events of the given occurrences as iCalendar
        string.

        """

        vcalendar = vCalendar()
        vcalendar.add('prodid', '-//OneGov//onegov.event//')
        vcalendar.add('version', '2.0')

        query = self.query().with_entities(Occurrence.event_id)
        event_ids = set([r.event_id for r in query])
        query = self.session.query(Event).filter(Event.id.in_(event_ids))
        for event in query:
            modified = event.modified or event.created or datetime.utcnow()

            vevent = vEvent()
            vevent.add('uid', f'{event.name}@onegov.event')
            vevent.add('summary', event.title)
            vevent.add('dtstart', to_timezone(event.start, UTC))
            vevent.add('dtend', to_timezone(event.end, UTC))
            vevent.add('last-modified', modified)
            vevent.add('dtstamp', modified)
            vevent.add('location', event.location)
            vevent.add('description', event.description)
            vevent.add('categories', event.tags)
            if event.recurrence:
                vevent.add('rrule', event.icalendar_recurrence)
            if event.coordinates:
                vevent.add(
                    'geo', (event.coordinates.lat, event.coordinates.lon)
                )
            vevent.add('url', request.link(event))
            vcalendar.add_component(vevent)

        return vcalendar.to_ical()



from click.testing import CliRunner
from datetime import datetime
from onegov.core.utils import module_path
from onegov.event.cli import cli
from onegov.event.collections import EventCollection
from onegov.event.models import Event
from os import path
from pytest import mark
from transaction import commit
from unittest.mock import MagicMock
from unittest.mock import patch


def test_import_ical(cfg_path, temporary_directory):
    runner = CliRunner()

    text = '\n'.join([
        'BEGIN:VCALENDAR',
        'VERSION:2.0',
        'PRODID:-//OneGov//onegov.event//',
        'BEGIN:VEVENT',
        'SUMMARY:Squirrel Park Visit',
        'UID:squirrel-park-visit@onegov.event',
        'DTSTART;VALUE=DATE-TIME:20150616T133000Z',
        'DTEND;VALUE=DATE-TIME:20150616T220000Z',
        'DTSTAMP;VALUE=DATE-TIME:20140101T000000Z',
        'RRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR,SA,SU;UNTIL=20150616T220000Z',
        'DESCRIPTION:<em>Furry</em> things will happen!',
        'CATEGORIES:fun',
        'CATEGORIES:animals',
        'LAST-MODIFIED;VALUE=DATE-TIME:20140101T000000Z',
        'LOCATION:Squirel Park',
        'GEO:47.051752750515746;8.305739625357093',
        'URL:https://example.org/event/squirrel-park-visit',
        'END:VEVENT',
        'END:VCALENDAR'
    ])
    ical = path.join(temporary_directory, 'ical.ics')
    with open(ical, 'w') as f:
        f.write(text)

    # First import
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'import-ical', ical
    ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "1 added, 0 updated, 0 deleted" in result.output

    # Re-import
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'import-ical', ical
    ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "0 added, 0 updated, 0 deleted" in result.output

    # Clear
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'clear'
    ], 'y')
    assert result.exit_code == 0

    # Import
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'import-ical', ical
    ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "1 added, 0 updated, 0 deleted" in result.output

    # Adjust ical
    with open(ical, 'w') as f:
        f.write(text.replace('Squirel', 'Squirrel'))

    # Reimport
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'import-ical', ical
    ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "0 added, 1 updated, 0 deleted" in result.output


@mark.parametrize("xml", [
    module_path('onegov.event', 'tests/fixtures/guidle.xml'),
])
def test_import_guidle(cfg_path, temporary_directory, xml):
    runner = CliRunner()

    with open(xml) as f:
        text = f.read()
    response = MagicMock(text=text)

    # First import
    with patch('onegov.event.cli.get', return_value=response):
        result = runner.invoke(cli, [
            '--config', cfg_path,
            '--select', '/foo/bar',
            'import-guidle', "'https://example.org/abcd'",
        ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "4 added, 0 updated, 0 deleted" in result.output

    # Reimport
    with patch('onegov.event.cli.get', return_value=response):
        result = runner.invoke(cli, [
            '--config', cfg_path,
            '--select', '/foo/bar',
            'import-guidle', "'https://example.org/abcd'",
        ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "0 added, 0 updated, 0 deleted" in result.output

    # Clear
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'clear'
    ], 'y')
    assert result.exit_code == 0

    # Import
    with patch('onegov.event.cli.get', return_value=response):
        result = runner.invoke(cli, [
            '--config', cfg_path,
            '--select', '/foo/bar',
            'import-guidle', "'https://example.org/abcd'",
        ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "4 added, 0 updated, 0 deleted" in result.output

    # Create tagmap
    tagmap = path.join(temporary_directory, 'tagmap.csv')
    with open(tagmap, 'w') as f:
        f.write("Konzert Pop / Rock / Jazz,Konzert\nSport,Sport")

    # Re-import with tagmap
    with patch('onegov.event.cli.get', return_value=response):
        result = runner.invoke(cli, [
            '--config', cfg_path,
            '--select', '/foo/bar',
            'import-guidle', "'https://example.org/abcd'",
            '--tagmap', tagmap
        ])
    assert result.exit_code == 0
    assert "Events successfully imported" in result.output
    assert "Tags not in tagmap: \"Kulinarik\"!"
    assert "0 added, 4 updated, 0 deleted" in result.output


def test_fetch(cfg_path, session_manager):
    runner = CliRunner()

    session_manager.ensure_schema_exists('foo-baz')
    session_manager.ensure_schema_exists('foo-qux')

    def get_session(entity):
        session_manager.set_current_schema(f'foo-{entity}')
        return session_manager.session()

    for entity, title, source, tags, location in (
        ('bar', '1', None, [], ''),
        ('bar', '2', None, ['A'], None),
        ('bar', '3', None, ['A', 'B'], 'bar'),
        ('bar', '4', None, ['A', 'C'], '1234 Bar'),
        ('bar', '5', None, ['C'], 'there in 4321 baz!'),
        ('bar', '6', 'xxx', [], 'bar'),
        ('bar', '7', 'yyy', ['A', 'B'], None),
        ('baz', 'a', None, [], 'BAZ'),
        ('baz', 'b', None, ['A', 'C'], '4321 Baz'),
        ('baz', 'c', 'zzz', ['B', 'C'], 'bar'),
    ):
        EventCollection(get_session(entity)).add(
            title=title,
            start=datetime(2015, 6, 16, 9, 30),
            end=datetime(2015, 6, 16, 18, 00),
            timezone='Europe/Zurich',
            tags=tags,
            location=location,
            source=source
        )
    commit()

    assert get_session('bar').query(Event).count() == 7
    assert get_session('baz').query(Event).count() == 3
    assert get_session('qux').query(Event).count() == 0

    # No sources provided
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/qux',
        'fetch',
    ])
    assert result.exit_code != 0
    assert "Provide at least one source" in result.output

    # Bar[*] -> Qux
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/qux',
        'fetch',
        '--source', 'bar'
    ])
    assert result.exit_code == 0
    assert "5 added, 0 updated, 0 deleted" in result.output

    # Bar[B, C] -> Qux
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/qux',
        'fetch',
        '--source', 'bar',
        '--tag', 'A',
        '--tag', 'B'
    ])
    assert result.exit_code == 0
    assert "0 added, 0 updated, 2 deleted" in result.output

    # Bar[C], Baz[C] -> Qux
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/qux',
        'fetch',
        '--source', 'bar',
        '--source', 'baz',
        '--tag', 'C',
    ])
    assert result.exit_code == 0
    assert "2 added, 0 updated, 2 deleted" in result.output

    # Baz['bar'] qux['bar'] -> Bar
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/bar',
        'fetch',
        '--source', 'baz',
        '--source', 'qux',
        '--location', 'bar',
    ])
    assert result.exit_code == 0
    assert "0 added, 0 updated, 0 deleted" in result.output

    # Bar['baz'] -> Baz
    result = runner.invoke(cli, [
        '--config', cfg_path,
        '--select', '/foo/baz',
        'fetch',
        '--source', 'qux',
        '--source', 'bar',
        '--location', 'baz',
    ])
    assert result.exit_code == 0
    assert "1 added, 0 updated, 0 deleted" in result.output



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.orm.types import JSON
from onegov.core.upgrade import upgrade_task
from onegov.event import EventCollection
from sqlalchemy import Column


@upgrade_task('Add coordinates column')
def add_coordinates_column(context):
    for table in ('events', 'event_occurrences'):
        context.operations.add_column(
            table, Column('coordinates', JSON(), nullable=True))


@upgrade_task('Drop coordinates column from occurrences')
def drop_coordinates_column(context):
    context.operations.drop_column('event_occurrences', 'coordinates')


@upgrade_task('Migrate coordinates column in events')
def migrate_coordinates_column(context):
    # merge the separate coordinates column into the content column
    # (gotta love postgres' json support!)
    context.session.execute("""
        UPDATE events
        SET "content" = (
            "content" || (
                '{"coordinates": ' || "coordinates"::text || '}'
            )::jsonb
        )
        WHERE coordinates IS NOT NULL
    """)

    context.operations.drop_column('events', 'coordinates')


@upgrade_task('Validate existing rrules')
def validate_existing_rrules(context):
    for event in EventCollection(context.session).query():
        event.validate_recurrence('recurrence', event.recurrence)



from onegov.core.orm.types import UTCDateTime
from sedate import to_timezone
from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import HSTORE
from sqlalchemy.ext.mutable import MutableDict


class OccurrenceMixin(object):
    """ Contains all attributes events and ocurrences share.

    The ``start`` and ``end`` date and times are stored in UTC - that is, they
    are stored internally without a timezone and are converted to UTC when
    getting or setting, see :class:`UTCDateTime`. Use the properties
    ``localized_start`` and ``localized_end`` to get the localized version of
    the date and times.
    """

    #: Title of the event
    title = Column(Text, nullable=False)

    #: A nice id for the url, readable by humans
    name = Column(Text)

    #: Description of the location of the event
    location = Column(Text, nullable=True)

    #: Tags/Categories of the event
    _tags = Column(MutableDict.as_mutable(HSTORE), nullable=True, name='tags')

    @property
    def tags(self):
        """ Tags/Categories of the event. """

        return list(self._tags.keys()) if self._tags else []

    @tags.setter
    def tags(self, value):
        self._tags = dict(((key.strip(), '') for key in value))

    #: Timezone of the event
    timezone = Column(String, nullable=False)

    #: Start date and time of the event (of the first event if recurring)
    start = Column(UTCDateTime, nullable=False)

    @property
    def localized_start(self):
        """ The localized version of the start date/time. """

        return to_timezone(self.start, self.timezone)

    #: End date and time of the event (of the first event if recurring)
    end = Column(UTCDateTime, nullable=False)

    @property
    def localized_end(self):
        """ The localized version of the end date/time. """

        return to_timezone(self.end, self.timezone)



import click
import pycurl

from csv import reader as csvreader
from dateutil.parser import parse
from hashlib import sha1
from io import BytesIO
from lxml import etree
from onegov.core.cli import abort
from onegov.core.cli import command_group
from onegov.core.cli import pass_group_context
from onegov.file import DepotApp
from onegov.event import log
from onegov.event.collections import EventCollection
from onegov.event.models import Event, EventFile
from onegov.event.models import Occurrence
from onegov.event.utils import GuidleExportData, as_rdates
from onegov.gis import Coordinates
from operator import add
from pathlib import Path
from requests import get
from sqlalchemy import or_
from sqlalchemy.dialects.postgresql import array

cli = command_group()


@cli.command('clear')
@pass_group_context
def clear(group_context):
    """ Deletes all events.

        onegov-event --select '/veranstaltungen/zug' clear

    """

    def _clear(request, app):
        if not click.confirm("Do you really want to remove all events?"):
            abort("Deletion process aborted")

        session = app.session()
        for event in session.query(Event):
            session.delete(event)
        for occurrence in session.query(Occurrence):
            session.delete(occurrence)

    return _clear


def download_image(image_url):
    buffer = BytesIO()

    c = pycurl.Curl()
    c.setopt(c.URL, image_url)
    c.setopt(c.WRITEDATA, buffer)
    c.setopt(c.FOLLOWLOCATION, True)

    try:
        c.perform()

        if c.getinfo(c.RESPONSE_CODE) != 200:
            return None

    finally:
        c.close()

    return buffer


@cli.command('import-json')
@pass_group_context
@click.argument('url')
@click.option('--tagmap', type=click.File())
def import_json(group_context, url, tagmap):
    """ Fetches the events from a seantis.dir.events instance.

    This command is intended for migration and to be removed in the future.

    Example:

        onegov-event --select '/veranstaltungen/zug' import-json \
        'https://veranstaltungen.zug.ch/veranstaltungen/?type=json&compact'

    """
    if tagmap:
        tagmap = {row[0]: row[1] for row in csvreader(tagmap)}

    def _import_json(request, app):
        unknown_tags = set()

        response = get(url)
        response.raise_for_status()
        response = response.json()

        session = app.session()
        events = EventCollection(session)

        for item in response:
            title = item['title']

            start = parse(item['start'])
            end = parse(item['end'])
            timezone = item['timezone']
            recurrence = item['recurrence']

            organizer = ', '.join((line for line in (
                item['organizer'] or '',
                item['contact_name'] or '',
                item['contact_email'] or '',
                item['contact_phone'] or ''
            ) if line))

            location = ', '.join((line for line in (
                item['locality'] or '',
                ' '.join((
                    item['street'] or '',
                    item['housenumber'] or ''
                )).strip(),
                ' '.join((
                    item['zipcode'] or '',
                    item['town'] or '',
                )).strip(),
            ) if line))

            description = '\r\n\r\n'.join((line for line in (
                item['short_description'] or '',
                item['long_description'] or '',
                item['event_url'] or '',
                item['location_url'] or '',
                item['prices'],
                item['registration']
            ) if line))

            tags = item['cat1']
            if tagmap and tags:
                unknown_tags |= set(tags) - tagmap.keys()
                tags = {tagmap[tag] for tag in tags if tag in tagmap}
            # todo: handle item['cat2']

            coordinates = None
            if item['latitude'] and item['longitude']:
                coordinates = Coordinates(
                    lat=item['latitude'],
                    lon=item['longitude']
                )

            event = Event(
                state='initiated',
                name=events._get_unique_name(title),
                title=title,
                start=start,
                end=end,
                timezone=timezone,
                description=description,
                organizer=organizer,
                location=location,
                coordinates=coordinates,
                tags=tags or [],
                meta={'submitter_email': item['submitter_email']},
            )

            # if the recurrence rule is not supported, turn it into a
            # list of rdates
            try:
                event.validate_recurrence('recurrence', recurrence)
            except RuntimeError:
                event.recurrence = as_rdates(recurrence)
            else:
                event.recurrence = recurrence

            if item['images']:
                buffer = download_image(item['images'][0]['url'])

                if buffer:
                    event.image = EventFile(
                        name=item['images'][0]['name'],
                        reference=buffer
                    )

            # todo: handle item['attachments']
            session.add(event)
            event.submit()
            event.publish()

        if unknown_tags:
            unknown_tags = ', '.join([f'"{tag}"' for tag in unknown_tags])
            click.secho(f"Tags not in tagmap: {unknown_tags}!", fg='yellow')

        click.secho(f"Imported {len(response)} events", fg='green')

    return _import_json


@cli.command('import-ical')
@pass_group_context
@click.argument('ical', type=click.File())
def import_ical(group_context, ical):
    """ Imports events from an iCalendar file.

    Example:

        onegov-event --select '/veranstaltungen/zug' import-ical import.ics

    """

    def _import_ical(request, app):
        collection = EventCollection(app.session())
        added, updated, purged = collection.from_ical(ical.read())
        click.secho(
            f"Events successfully imported "
            f"({added} added, {updated} updated, {purged} deleted)",
            fg='green')

    return _import_ical


@cli.command('import-guidle')
@pass_group_context
@click.argument('url')
@click.option('--tagmap', type=click.File())
def import_guidle(group_context, url, tagmap):
    """ Fetches the events from guidle.

    Example:

        onegov-event --select '/veranstaltungen/zug' import-guidle \
        'http://www.guidle.com/xxxx/'

    """
    if tagmap:
        tagmap = {row[0]: row[1] for row in csvreader(tagmap)}

    def _import_guidle(request, app):
        try:
            response = get(url)
            response.raise_for_status()

            unknown_tags = set()
            prefix = 'guidle-{}'.format(sha1(url.encode()).hexdigest()[:10])

            events = []
            root = etree.fromstring(response.text.encode('utf-8'))
            for offer in GuidleExportData(root).offers():
                tags, unknown = offer.tags(tagmap)
                unknown_tags |= unknown

                # medium sized images seem to be larger than what we actually
                # use, so no need to use bigger sizes (if we do, we need to
                # change the download_image function to not operate solely
                # in memory)
                if isinstance(app, DepotApp):
                    image_url = offer.image_url(size='medium')
                else:
                    image_url = None

                if image_url:
                    buffer = download_image(image_url)
                else:
                    buffer = None

                for index, schedule in enumerate(offer.schedules()):
                    event = Event(
                        state='initiated',
                        title=offer.title,
                        start=schedule.start,
                        end=schedule.end,
                        recurrence=schedule.recurrence,
                        timezone=schedule.timezone,
                        description=offer.description,
                        organizer=offer.organizer,
                        location=offer.location,
                        coordinates=offer.coordinates,
                        tags=tags,
                        source=f'{prefix}-{offer.uid}.{index}',
                    )

                    if buffer:
                        buffer.seek(0)

                        event.image = EventFile(
                            name=image_url.rsplit('/', 1)[-1],
                            reference=buffer
                        )

                    events.append(event)

                # todo: handle attachements

            collection = EventCollection(app.session())
            added, updated, purged = collection.from_import(events, prefix)

            if unknown_tags:
                unknown_tags = ', '.join([f'"{tag}"' for tag in unknown_tags])
                click.secho(
                    f"Tags not in tagmap: {unknown_tags}!", fg='yellow'
                )

            click.secho(
                f"Events successfully imported from '{url}' "
                f"({added} added, {updated} updated, {purged} deleted)",
                fg='green'
            )
        except Exception as e:
            log.error("Error importing events", exc_info=True)
            raise(e)

    return _import_guidle


@cli.command()
@pass_group_context
@click.option('--source', multiple=True)
@click.option('--tag', multiple=True)
@click.option('--location', multiple=True)
def fetch(group_context, source, tag, location):
    """ Fetches events from other instances.

    Only fetches events from the same namespace which have not been imported
    themselves.

    Example

        onegov-event --select '/veranstaltungen/zug' fetch \
            --source menzingen --source steinhausen
            --tag Sport --tag Konzert

    """

    def vector_add(a, b):
        return list(map(add, a, b))

    if not len(source):
        abort("Provide at least one source")

    def _fetch(request, app):

        # XXX use a proper depot manager for this
        def event_file_path(reference):
            return Path(app.depot_storage_path) / reference['path'] / 'file'

        try:
            result = [0, 0, 0]

            for key in source:
                remote_schema = '{}-{}'.format(app.namespace, key)
                local_schema = app.session_manager.current_schema
                assert remote_schema in app.session_manager.list_schemas()

                app.session_manager.set_current_schema(remote_schema)
                remote_session = app.session_manager.session()
                assert remote_session.info['schema'] == remote_schema

                query = remote_session.query(Event)
                query = query.filter(
                    or_(
                        Event.meta['source'].astext.is_(None),
                        Event.meta['source'].astext == ''
                    )
                )
                if tag:
                    query = query.filter(Event._tags.has_any(array(tag)))
                if location:
                    query = query.filter(
                        or_(*[
                            Event.location.ilike(f'%{term}%')
                            for term in location
                        ])
                    )

                remote_events = [
                    Event(
                        state='initiated',
                        title=event.title,
                        start=event.start,
                        end=event.end,
                        timezone=event.timezone,
                        recurrence=event.recurrence,
                        content=event.content,
                        location=event.location,
                        tags=event.tags,
                        image=(
                            event.image and
                            EventFile(
                                name=event.image.name,
                                reference=event_file_path(
                                    event.image.reference
                                ).open('rb')
                            ) or None
                        ),
                        source=f'fetch-{key}-{event.name}',
                        coordinates=event.coordinates,
                    ) for event in query
                ]

                # be sure to switch back to the local schema, lest we
                # accidentally update things on the remote
                app.session_manager.set_current_schema(local_schema)
                local_events = EventCollection(app.session_manager.session())

                result = vector_add(
                    result,
                    local_events.from_import(remote_events, f'fetch-{key}')
                )
                # todo: attachments

            click.secho(
                f"Events successfully fetched "
                f"({result[0]} added, {result[1]} updated, "
                f"{result[2]} deleted)",
                fg='green'
            )

        except Exception as e:
            log.error("Error fetching events", exc_info=True)
            raise(e)

    return _fetch
